"""From peak energies to the Periodic Table"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/03_moseley-law.ipynb.

# %% auto 0
__all__ = ['elementsInstance', 'make_ptable', 'ptable_plot', 'XFluo', 'moseley_law', 'moseley_plot']

# %% ../notebooks/03_moseley-law.ipynb 21
# Periodic Table 
import mendeleev as mv 
from mendeleev.fetch import fetch_table 

# basic XRF physics 
from fisx import Elements 

# advanced XRF stuff that we do no need now   
#from fisx import Material
#from fisx import Detector
#from fisx import XRF 

# peak finding 
import scipy.signal as ssg
import scipy

# plotting 
import pandas as pd 
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches 
import matplotlib.cm as cm 
import seaborn as sb

# %% ../notebooks/03_moseley-law.ipynb 22
def make_ptable(): 
    '''Make numpy array with element attributes for regular part of the periodic table.
    
    The irregular Lanthanides and Actinides series are rare, so we do not plot them'''
    
    #ptable_df = mv.get_table('elements') # mendeleev 0.5 
    ptable_df = fetch_table('elements')
    ptable = ptable_df[['atomic_number', 'symbol', 'name', 'group_id', 'period']].values
    is_regular = ~np.isnan(ptable_df['group_id'].values)
    
    return ptable, is_regular


def _colorize(elem_select, crop=True): 
    '''High contrast standard color palette for list of elements *elem_select*.
    
    '''
    
    DEFAULT_COLOR = [0.4, 0.4, 0.9, 0.3] 
    
    # return single color tuple is argument is string, e.g. 'Fe' 
    return_single_color_tuple = False 
    if type(elem_select) == str:
        return_single_color_tuple = True 
        elem_select = [elem_select]
        
    # create standard palette for Hydrogen(Z=1) to Organesson(Z=117) 
    palette = []
    for tone in ['pastel', 'bright', 'deep', 'dark'] * 3:  
        palette.extend(sb.palettes.color_palette(tone))
    palette = np.array(palette)[0:118]
    # and include unity alpha channel
    palette = np.c_[palette, np.ones(len(palette))]
    
    # return full color palette if argument is empty list 
    if len(elem_select) == 0:
        return palette
 
    else: 
        # find index position (Z-1) of elements   
        ptable_indices = [mv.element(e).atomic_number - 1 for e in elem_select]
        
        # initialize all element color list with light blue grey 
        colors = np.zeros([118, 4])
        colors[:] = DEFAULT_COLOR 
        
        # colorize selected elements with standard palette color 
        colors[ptable_indices] = palette[ptable_indices]
        
        # skip colors for non selected elements 
        if crop: 
            colors = palette[ptable_indices]

            # for single element just return color tuple 
            if return_single_color_tuple: 
                return colors[0]   

            else: 
                return colors, ptable_indices 
            
        return colors, ptable_indices


    
def _draw_box(ax, element_attrs, edgecolor=None, facecolor=None):
    '''Draw a box for element *element_attrs*  in subplot *ax*. '''
    
    w = 0.9
    h = 0.9
    
    atomic_number, symbol, name, x, y = element_attrs 
    
    if edgecolor is None: 
        edgecolor = [0.8, 0.8, 1.0]
        
    if facecolor is None: 
        facecolor = [0.95, 0.95, 1.0]
    
    rect = mpatches.Rectangle([x - w/2, y - h/2], h, w, edgecolor=edgecolor, facecolor=facecolor)
    ax.add_patch(rect)
    
    ax.annotate(symbol, [x, y], xytext=[0, -4], textcoords='offset points',
                va='center', ha='center', fontsize=17)
    
    ax.annotate(atomic_number, [x, y], xytext=[0, 12], textcoords='offset points', 
                va='center', ha='center', fontsize=10)
    
    ax.annotate(name, [x, y], xytext=[0, -17], textcoords='offset points',
                va='center', ha='center', fontsize=5)
    

def ptable_plot(elem_select=None, figname=None): 
    '''Create periodic table plot with selected elements colorized.'''
    
    if elem_select is None: 
        elem_select = []
    
    # initialize full periodic table 
    ptable, is_regular = make_ptable() 
    
    # colorize full table 
    colors, _ = _colorize(elem_select, crop=False) 
    
    # continue with regular elements only  
    ptable = ptable[is_regular] 
    colors = colors[is_regular]
    
    # create figure 
    fig, ax = plt.subplots(figsize=[14, 8])
    ax.invert_yaxis()

    for i, element_attrs in enumerate(ptable):
        
        _draw_box(ax, element_attrs, facecolor=colors[i])

    ax.set_xlim(0, 19)
    ax.set_ylim(8, 0)
    ax.axis('off');
    
    if figname is not None: 

        fig.savefig(figname) 
        


# fisx instantiation is needed before continuing 
elementsInstance = Elements()
elementsInstance.initializeAsPyMca()

class XFluo: 
    
    def __init__(self, element, tube_keV, weight_list='equal', std=0.01, min_prom=0.001): 
        '''Create an annotated x-ray fluorescence spectrum object. 
        
        Args: 
            element (str): Chemical symbol e.g 'Fe' 
            tube_keV (float or list of floats): X-ray tube energies in keV. 
               For a poly chromatic X-ray tube multiple energies can be provided.  
            weight_list (list of numbers): X-ray energy weights. 
                Corresponding to energy_list     
            std (number): Peak width used for plotting 
            min_prom (number): Minimal peak prominance. 
                Only peaks higher then prominance are labeled. 
                
        Example: 
            xf = XFluo('Fe', [20, 40])
                
        
        Returns: 
            XRF annotated spectrum object 
        '''
        # convert into list if single number
        energy_list = np.array([tube_keV]).flatten() 
        
        sorting = np.argsort(energy_list)
        energy_list = np.array(energy_list)[sorting] # need to make sure last energy is highest ? 
        
        if weight_list == 'equal': 
            weight_list = np.ones_like(energy_list) 
            
        weight_list = np.array(weight_list)[sorting]
        
        self.element = element 
        self.atomic_number = mv.element(element).atomic_number 
            
        self.x = np.linspace(0, max(energy_list), 2000) 
        
        self.total_spectrum_stack = np.zeros([len(energy_list), len(self.x)]) # initialize 

        # in order to get all lines in lines_dict need sorted energy list 
        for j, energy in enumerate(energy_list): 
        
            self.lines_dict = elementsInstance.getExcitationFactors(element, [energy]) 

            self.lines_df = pd.DataFrame(self.lines_dict).T 

            self.energies = self.lines_df['energy'].values 
            self.excitations = self.lines_df.index.values 
            self.rates = self.lines_df['rate'].values 
            #self.factors = self.lines_df['factor'].values # unclear what factor is 

            self.line_spectra = np.zeros([len(self.energies), len(self.x)]) # initialize temporary stack 

            for i, energy in enumerate(self.energies): 
                self.line_spectra[i] = self.rates[i] * np.exp(-(1 / std) * (self.x - energy)**2)

            self.total_spectrum_stack[j] = weight_list[j] * self.line_spectra.sum(axis=0) # combine gaussians  
        
        self.total_spectrum = self.total_spectrum_stack.sum(axis=0) 
        
        # normalizing heighest peak to unity 
        self.total_spectrum = self.total_spectrum / self.total_spectrum.max()
        
        peaks, props = ssg.find_peaks(self.total_spectrum, prominence=[min_prom, None])
        
        self.peak_energies = self.x[peaks]
        self.peak_intensities = self.total_spectrum[peaks] 
        
        # find indices of nearest excitations 
        # corrected bug with .values 
        nearest = np.argmin((self.lines_df['energy'].values[:, None] - self.peak_energies[None, :])**2, axis=0)  
        self.peak_excits = list(self.lines_df.index[nearest])
        
        
        
    def plot(self, ax=None, color=None, figsize=[12, 5], tight_layout=True, up=False, mos=False, peak_labels='simple'):
        '''Plot an xrf spectrum'''

        if ax is None: 
            fig, ax = plt.subplots(figsize=figsize) 
        else: 
            fig = ax.get_figure() 
            
        if up is False: 
            Z = 0 
            if mos is False: 
                h = 1 
            elif mos is True: 
                h = 4 
        else: 
            Z = self.atomic_number 
            h = 4 
            
        x = self.x 
        y = h * self.total_spectrum + Z
        
        px = self.peak_energies
        py = h * self.peak_intensities + Z 

        #ax.fill_between(self.x, self.total_spectrum, color=color, alpha=0.3)
        #ax.plot(self.x, self.total_spectrum, color=color)
        #ax.scatter(self.peak_energies, self.peak_intensities + dy, color=color) 
        
        if color is None: 
            color = _colorize(self.element)
        
        ax.fill_between(x, y, y2=Z, color=color, alpha=0.3)
        ax.plot(x, y, color=color)
        
        if peak_labels != 'none': 
            ax.scatter(px, py, color=color)
            for e in px: 
                ax.axvline(e, color=color, alpha=0.3)
        
        ax.set_xlabel('energy (keV)')
        #ax.set_ylim(0, 1.2)
        
        
        # peak labels 
        if peak_labels == 'full': 
            xy_list = list(zip(px, py))
            for i, s in enumerate(self.peak_excits): 
                ax.annotate(f'{self.element}\n{s}\n{self.peak_energies[i]:.3f}keV',
                            xy_list[i], 
                            xycoords='data', xytext=(0, 7), textcoords='offset points',
                            color=color, 
                            bbox=dict(boxstyle="round", alpha=0, edgecolor='white', facecolor='white')) 
                
        elif peak_labels == 'simple': 
            xy_list = list(zip(px, py))
            for i, s in enumerate(self.peak_excits): 
                ax.annotate(f'{self.element}',
                            xy_list[i], 
                            xycoords='data', xytext=(0, 7), textcoords='offset points',
                            color=color, 
                            bbox=dict(boxstyle="round", alpha=0, edgecolor='white', facecolor='white'))   
                
        elif peak_labels == 'none': 
            pass
        
        # seems to slow down plotting 
        #or not? 
        if tight_layout:
            fig.tight_layout()
            
        # (self.peak_energies[i], self.peak_intensities[i]),           

def moseley_law(E_K_alpha_keV): 
    '''Square root form of Moseley's law. 
    
    Args: 
        E_K_alfa_keV (float or array of floats): K_alpha peak energy in keV 
        
    Example: 
        moseley_law(6.40)
        
    Returns: 
        Z (float or array of floats): predicted atomic number 
        
    ''' 
    Z = 1 + np.sqrt(1000 * E_K_alpha_keV / 10.2) 
    
    return Z 


def moseley_plot(tube_keV, elem_select=None, weight_list='equal', law=True, figname=None): 
    '''Generate Moseley plot of simulated xrf spectra for periodic table.
    
    Args: 
        elem_select (list of str): List of symbols of elements. 
            All selected elements will be highlighted with bright colors and peak labels. 
        tube_keV (float or list of floats): X-ray tube energy in keV. 
        weight_list (list of floats): X-ray tube intensities. 
            Optional, defaults to 'equal'. 
        
    Example: 
        moseley_plot(tube_keV=40, elem_select=['Fe', 'Ca']);
    
    Returns: 
        matplotlib figure 
            
    '''


    if elem_select is None: 
        elem_select = []

    # initialize full periodic table 
    ptable, is_regular = make_ptable() 

    # colorize full table 
    colors, indices_selected = _colorize(elem_select, crop=False) 

    atomic_numbers, symbols, names = ptable.T[0:3]

    fig, ax = plt.subplots(figsize=[12, 8])
    
    E_keV_list = np.linspace(0, np.max(tube_keV), 2000) 
    Z = moseley_law(E_keV_list) 
    
    ax.plot(E_keV_list, Z, zorder=-10, linestyle=':', color='k', label="Moseley's law")

    Pb_Si = symbols[13:82] [::-1]
    Pb_Si_colors = colors[13:82] [::-1] 

    for s, c in list(zip(Pb_Si, Pb_Si_colors)): 
        if s in elem_select: 
            peak_labels = 'simple' 

            # ideal spectrum 
            XFluo(s, tube_keV, weight_list=weight_list).plot(ax=ax, color=c, up=False, mos=True, 
                                                             peak_labels=peak_labels)

        else: 
            peak_labels = 'none'

        XFluo(s, tube_keV, weight_list=weight_list).plot(ax=ax, color=c, up=True, peak_labels=peak_labels)

    #ax.set_yticks(atomic_numbers)
    #ax.set_yticklabels(tick_labels)

    ax.set_ylabel('Atomic number')
    ax.set_xticks(range(int(tube_keV)))
    ax.set_xlim(0, np.max(tube_keV))
    ax.set_ylim(0, 90)
    ax.grid(False) 
    ax.set_title(f'Moseley plot (x-ray tube at {np.max(tube_keV)} keV)')
    ax.legend()
    plt.tight_layout()
    
    if figname is not None: 
        print(f'Saving plot as: {figname}...')
        fig.savefig(figname)
        
    return fig



"""The forest and the trees"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/02_a-peak-pattern-atlas.ipynb.

# %% auto 0
__all__ = ['PeakPatternAtlas', 'colorize', 'PeriodicTable']

# %% ../notebooks/02_a-peak-pattern-atlas.ipynb 11
# Periodic Table 
import mendeleev as mv 
from mendeleev.fetch import fetch_table  

import moseley as mos 
from . import ElementXRF

# plotting 
import pandas as pd 
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches 
import matplotlib.colors as mcolors 

import holoviews as hv  


# %% ../notebooks/02_a-peak-pattern-atlas.ipynb 12
class PeakPatternAtlas():
    '''Create a PeakPatternAtlas instance. '''

    def __init__(self, EOI=None, excitation_energy_keV=25, x_keVs=None, verbose=True, order='alpha'):
        '''Compute an ordered Peak Pattern Atlas for `excitation_energy_keV=25`.'''

        self.table = mos.PeriodicTable(EOI=EOI)

        self.EOI = [element for element in self.table.EOI if not '#'in element]
        
        self.element_xrf_list = [] 
        self.ptrn_dict_list = []

        for i, element in enumerate(self.EOI):     
            if verbose: 
                print(f'Please wait while computing spectral pattern for element {i+1}/{len(self.EOI)}...', end='\r') 

            element_xrf = ElementXRF(element, excitation_energy_keV=excitation_energy_keV, x_keVs=x_keVs)
            ptrn_dict = element_xrf.ptrn_dict

            self.element_xrf_list.append(element_xrf)
            self.ptrn_dict_list.append(ptrn_dict)

        if order == 'alpha':
            alphas = [ptrn['alpha_keV'] for ptrn in self.ptrn_dict_list]
            idxs = np.argsort(alphas)
            
        elif order == 'Z': 
            atomic_numbers = [ptrn['atomic_number'] for ptrn in self.ptrn_dict_list] 
            idxs = np.argsort(atomic_numbers) 

        self.ptrn_dict_list = [self.ptrn_dict_list[idx] for idx in idxs]  
        self.element_xrf_list = [self.element_xrf_list[idx] for idx in idxs]  
        self.yticks = [(i, xrf.element) for i, xrf in enumerate(self.element_xrf_list)]
            
        if verbose: 
            print('Ready building Peak Pattern Atlas!                                   ') 

    def plot_atlas(self, moseley=False): 
        '''Build a holoviews peak pattern atlas. '''
    
        # standard element colors
        ptable = mos.PeriodicTable()
        color_dict = mos.colorize(ptable.ptable_regular)
        
        scatters = []
        lines = []
        yticks = []
        alphas = []
        
        Energy = hv.Dimension('Energy', unit='keV', range=(0, 20))
        
        for i, ptrn in enumerate(self.ptrn_dict_list): 
            # we should avoid recalculation here and use the ppa instance 
        
            element = ptrn['elem'] 
            Z = ptrn['atomic_number'] 
            color = color_dict[element]
            
            if moseley: 
                y = Z 
                color='grey'
            else: 
                y = i
                
            peaks_x, peaks_y = ptrn['peaks_xy'].T 
            x_min = float(peaks_x.min()) 
            x_max =float(peaks_x.max())
            ptrn_y = np.ones_like(peaks_y) * y
            
            alpha_keV = ptrn['alpha_keV'] 
                    
            scatters.append(hv.Scatter((peaks_x, ptrn_y), Energy, 'Element').opts(size=5, color=color))
            lines.append(hv.Curve([(x_min, y), (x_max, y)], Energy, 'Element').opts(color=color)) 
            alphas.append(hv.Scatter(([alpha_keV], [y]), Energy, 'Element').opts(size=7, marker='square', color=color)) 
        
            yticks.append((y, element))
        
        atlas = (hv.Overlay(lines) * hv.Overlay(scatters) * hv.Overlay(alphas))
        atlas = atlas.opts(yticks=yticks, frame_height=500, frame_width=500, xlim=(0, 22))
    
        return atlas
      


def _fetch_table(): 
    '''Utility function wrapper for mendeleev periodic table data.  
    
    Make dataframe with selected element attributes for regular part of the periodic table.
    The irregular Lanthanides and Actinides series are rare, so we do not plot them. 
    
    Returns: ptable_regular (dataframe) 
    '''
    
    #ptable_df = mv.get_table('elements') # mendeleev 0.5 
    ptable_df = fetch_table('elements')
    ptable = ptable_df[['atomic_number', 'symbol', 'name', 'group_id', 'period']]
    is_regular = ~np.isnan(ptable_df['group_id'].values)

    ptable_regular = ptable[is_regular]
    
    return ptable_regular 


def colorize(ptable_regular): 
    '''Generate regular element colors dictionary. 
    
    Returns: element_colors_dict '''
    
    n_regular = len(ptable_regular) # 90 rows 
    
    # see: https://matplotlib.org/stable/gallery/color/named_colors.html 
    css4_dict = mcolors.CSS4_COLORS 
        
    # hack to select brightest colors and exclude american double grays 
    css4_names = list(css4_dict.keys()) 
    css4_names = sorted(sorted(css4_names, key=lambda c: tuple(mcolors.rgb_to_hsv(mcolors.to_rgb(c))))) # brightest colors 
    css4_names = [name for name in css4_names if 'gray' not in name] # exclude doubles grays 

    # pick brightest 90 
    n_names = len(css4_names) 
    css4_names = css4_names[n_names - n_regular - 1:n_names - 1]

    element_colors_dict = dict(zip(ptable_regular['symbol'], css4_names))

    return element_colors_dict


def _draw_box(ax, element_attrs, edgecolor=None, facecolor=None):
    '''Draw a box for element *element_attrs*  in subplot *ax*. '''
    
    w = 0.9
    h = 0.9
    
    atomic_number, symbol, name, x, y = element_attrs 
    
    if edgecolor is None: 
        edgecolor = [0.8, 0.8, 1.0]
        
    if facecolor is None: 
        facecolor = [0.95, 0.95, 1.0]
    
    rect = mpatches.Rectangle([x - w/2, y - h/2], h, w, edgecolor=edgecolor, facecolor=facecolor)
    ax.add_patch(rect)
    
    ax.annotate(symbol, [x, y], xytext=[0, -4], textcoords='offset points',
                va='center', ha='center', fontsize=17)
    
    ax.annotate(atomic_number, [x, y], xytext=[0, 12], textcoords='offset points', 
                va='center', ha='center', fontsize=10)
    
    ax.annotate(name, [x, y], xytext=[0, -17], textcoords='offset points',
                va='center', ha='center', fontsize=5)


def _sort_ptrn_list(ptrn_dict_list, order='alpha'): 
    '''Utility function to sort ptrn_dict_list in 'alpha' or 'Z' order. 
    
    Returns: sorted_ptrn_list '''

    if order == 'alpha': 
        alpha_list = [ptrn['alpha_keV'] for ptrn in ptrn_dict_list]
        idxs = np.argsort(alpha_list)
    elif order == 'Z': 
        Z_list = [ptrn['atomic_number'] for ptrn in ptrn_dict_list] 
        idxs = np.argsort(Z_list)
        
    sorted_ptrn_list = [ptrn_dict_list[i] for i in idxs]  
        
    return sorted_ptrn_list


class PeriodicTable(): 
    '''Create a periodic table of elements. '''
    
    
    def __init__(self, EOI=None): 

        if EOI is None: 
            self.EOI = ['#H', '#He', '#Li', '#Be', '#B', '#C', 'N', 'O', 'F', 'Ne', 'Na', 'Mg', 'Al', 
                        'Si', 'P', 'S', 'Cl', 'Ar', 'K', 'Ca', '#Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe', 
                        'Co', 'Ni', 'Cu', 'Zn', '#Ga', '#Ge', 'As', '#Se', 'Br', '#Kr', '#Rb', 'Sr', 
                        '#Y', '#Zr', '#Nb', '#Mo', '#Tc', '#Ru', 'Rh', '#Pd', 'Ag', 'Cd', '#In', 'Sn', 
                        '#Sb', '#Te', 'I', '#Xe', '#Cs', 'Ba', '#La', '#Hf', '#Ta', '#W', '#Re', '#Os', 
                        '#Ir', '#Pt', 'Au', 'Hg', '#Tl', 'Pb', '#Bi', '#Po', '#At', '#Rn', '#Fr', '#Ra', 
                        '#Ac', '#Rf', '#Db', '#Sg', '#Bh', '#Hs', '#Mt', '#Ds', '#Rg', '#Cn', '#Nh', 
                        '#Fl', '#Mc', '#Lv', '#Ts', '#Og']    
        else: 
            self.EOI = EOI 
     
    
        # fetch regular part of periodic table 
        self.ptable_regular = _fetch_table() 
    
        # generate element colors dict 
        self.element_colors_dict = colorize(self.ptable_regular) 


    def plot_table(self, ax=None, figname=None): 
        '''Create periodic table plot with selected elements colorized.'''
    
        if ax is None: 
            fig, ax = plt.subplots(figsize=[14, 8])
        else: 
            fig = ax.get_figure()
            
        ax.invert_yaxis()
    
        for i, element_attrs in enumerate(self.ptable_regular.values): 
    
            atomic_number, symbol, name, x, y = element_attrs 
            facecolor = self.element_colors_dict[symbol] 
            
            if symbol in self.EOI: 
                edgecolor = 'r' 
            else: 
                edgecolor = None
            _draw_box(ax, element_attrs, edgecolor=edgecolor, facecolor=facecolor)
    
        ax.set_xlim(0, 19)
        ax.set_ylim(8, 0)
        ax.axis('off');
        
        if figname is not None: 
    
            fig.savefig(figname)
    
        return ax



"""Predicting elemental XRF (X-Ray Fluorescence) spectra"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/01_theoretical-peak-patterns.ipynb.

# %% auto 0
__all__ = ['ElementXRF', 'gaussian_convolve', 'find_peaks', 'plot_pattern', 'get_attenuation']

# %% ../notebooks/01_theoretical-peak-patterns.ipynb 9
import numpy as np 
import xraydb 
import pandas as pd 
import scipy.signal as ssg 
import scipy.interpolate as sip
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import mendeleev

# %% ../notebooks/01_theoretical-peak-patterns.ipynb 10
class ElementXRF(): 
    '''Computes fluorescence emission line energies and intensities for `element`. 
    
    '''
    
    def __init__(self, element, excitation_energy_keV='rhodium_Ka', h_mm=0.1, x_keVs=None, std=0.01, min_prom=0.001): 
        '''Create a theoretical elemental xrf spectrum object. '''

        ### LINES ### 

        if excitation_energy_keV == 'rhodium_Ka': 
            excitation_energy_keV = 22.1  

        self.element = element 

        excitation_energy = 1000 * excitation_energy_keV # keV -> eV 

        lines = xraydb.xray_lines(element, excitation_energy=excitation_energy) 

        line_names = [] # Siegbahn notation
        line_labels = [] # IUPAC notation 
        line_energies = [] 
        line_probabilities = [] 

        line_fluo_yields = []
        line_jump_coeffs = [] 
        line_attenuations = [] 
        
        line_intensities = []
        
        for name, line in lines.items(): 

            # xraydb 'intensities' (a.k.a. transition probablities) sum up to unity within each level 
            emission_energy, probability, initial_level, final_level = line  
            emission_energy_keV = emission_energy / 1000  
            label = f'{element}_{initial_level}{final_level}' #  construct IUPAC notation 
            
            # get corresponding edge properties 
            edge = initial_level # IUPAC notation!  e.g. 'L1', not 'La'
            edge_energy, fluo_yield, jump_ratio = xraydb.xray_edge(element, edge) 
            jump_coeff = (jump_ratio - 1) / jump_ratio # see Volker Thomsen 

            attenuation = get_attenuation(element, emission_energy_keV, excitation_energy_keV, h_mm=h_mm)

            # multiplying edge jump coefficient, transition probability and fluorescence yield... 
            intensity = jump_coeff * probability * fluo_yield * attenuation

            line_names.append(name) 
            line_labels.append(label)
            line_energies.append(emission_energy_keV)
            line_probabilities.append(probability)  
            
            line_fluo_yields.append(fluo_yield)
            line_jump_coeffs.append(jump_coeff) 
            line_attenuations.append(attenuation) 
            
            line_intensities.append(intensity) 

        self.lines_table = pd.DataFrame({'name': line_names, 
                                         'label': line_labels, 
                                         'energy': line_energies, 
                                         'trans_prob': line_probabilities, 
                                         'fluo_yield': line_fluo_yields, 
                                         'jump_coeff': line_jump_coeffs, 
                                         'attenuation': line_attenuations, 
                                         'intensity': line_intensities}) 

    
        ### CONTINUOUS ### 
        
        self.spectrum_xy = gaussian_convolve(line_energies, line_intensities, x_keVs=x_keVs, std=std) 
        self.peaks_xy = find_peaks(*self.spectrum_xy, min_prom=min_prom) 

        # sort high to low (just in case) 
        sort_idxs = np.argsort(self.peaks_xy[:, 1])[::-1]
        self.peaks_xy = self.peaks_xy[sort_idxs]

        # find nearest line Siegbahn peak labels 
        peaks_x, peaks_y = self.peaks_xy.T
        
        self.peak_labels = []
        for peak_keV in peaks_x:     
            line_idx = int(np.argmin((self.lines_table['energy'].values   - peak_keV)**2))
            line_label = self.lines_table['name'].values[line_idx]
            self.peak_labels.append(line_label)

        # create element pattern dict 

        alpha_keV = float(self.peaks_xy[0][0])
        alpha_escape_keV = alpha_keV - 1.74  # Silicon detector escape energy shift 
        
        # element   
        elem = self.element
        name = mendeleev.element(elem).name
        atomic_number = mendeleev.element(elem).atomic_number
        
        self.ptrn_dict = {
            'elem': elem,  
            'atomic_number': atomic_number, 
            'alpha_keV': alpha_keV, 
            'name': name, 
            'peaks_xy': self.peaks_xy, 
            'peak_labels': self.peak_labels, 
            'alpha_escape_keV': alpha_escape_keV} 
        
        

    def plot_spectrum(self, ax=None, edgecolor=None, facecolor=None, vlines_colors=None, xlim=[0, 25], peak_labels=True): 
        
        if edgecolor is None: 
            edgecolor = cm.tab20(0)
        if facecolor is None: 
            facecolor = 'w' 
    
        if ax is None: 
            fig, ax = plt.subplots() 
    
        x, y = self.spectrum_xy 
        peaks_x, peaks_y = self.peaks_xy.T
    
        ax.plot(x, y, color=edgecolor, zorder=9)
        ax.scatter(peaks_x[0], peaks_y[0], marker='s', facecolor=facecolor, edgecolor=edgecolor, zorder=10)  
        ax.scatter(peaks_x[1:], peaks_y[1:], marker='o', facecolor=facecolor, edgecolor=edgecolor, zorder=10)

        # add peak labels
        if peak_labels: 
            for label, xy in zip(self.peak_labels, self.peaks_xy): 
                ax.annotate(label, xy, xytext=(0, 5), textcoords="offset points",
                            va="bottom", ha="center") 
    
        # need to interpolate spectrum to add vlines 
    
        spectrum = sip.interp1d(x, y)
        line_keVs = self.lines_table['energy'] 
    
        line_heights = spectrum(line_keVs)
    
        ax.vlines(line_keVs, np.zeros_like(line_keVs), line_heights, colors=vlines_colors, alpha=0.5, zorder=8) 

        ax.set_title(f'Theoretical X-Ray Fluorescence for {self.element}') 
        ax.set_xlabel('Energy (keV)')
        ax.set_xlim(xlim)
        ax.set_ylim(ymax=1.1 * y.max())
        
        return ax 



def gaussian_convolve(peak_energies, peak_intensities, x_keVs=None, std=0.01): 
    '''Convolves line spectrum defined by `peak_energies` and `peak_intensities` 
    
    with a Gaussian peak shape.'''
    
    if x_keVs is None: 
        x_keVs = np.linspace(0, 40, 10000)

    y_spectrum = np.zeros_like(x_keVs) 

    for peak_energy, peak_intensity in zip(peak_energies, peak_intensities): 

        y_spectrum += peak_intensity * np.exp(-(1 / std) * (x_keVs - peak_energy)**2)
        
    return x_keVs, y_spectrum  

def find_peaks(x, y, min_prom=0.001):  
    '''Find peaks in spectrum `x`, `y` with minimal prominence `min_prom=`
    
    Returns: peaks_xy''' 
    
    # find the peaks (channel indexes) 
    peaks, props = ssg.find_peaks(y/y.max(), prominence=[min_prom, None])

    peaks_x = x[peaks]
    peaks_y = y[peaks] 
    
    # sort according to peak_height 
    sort_idxs = np.argsort(peaks_y)[::-1] 

    peaks_x = peaks_x[sort_idxs]
    peaks_y = peaks_y[sort_idxs]
    
    peaks_xy = np.c_[peaks_x, peaks_y]

    return peaks_xy 

def plot_pattern(ptrn_dict, ax=None, offset=0, color=None): 
    '''Plot a single element peak pattern. 
    ''' 
        
    element = ptrn_dict['elem']
    
    x, _ = ptrn_dict['peaks_xy'].T
    y = np.zeros_like(x) + offset 
    alpha_escape_keV = ptrn_dict['alpha_escape_keV']
         
    left_x = min(x)
    right_x = max(x) 

    if color is None: 
        facecolor = 'w'
        edgecolor = 'grey'
        linecolor = 'grey'
        linestyle = ':'

    else: 
        facecolor = color 
        edgecolor = 'k'
        linecolor = color
        linestyle = '-'

    if ax is None: 
        fig, ax = plt.subplots() 

    ax.scatter(x, y, s=15, facecolor=facecolor, edgecolor=edgecolor)
    ax.scatter(x[0], y[0], marker='s', facecolor=facecolor, edgecolor=edgecolor)
    ax.plot([left_x, right_x], [y[0], y[0]], color=edgecolor, linestyle=linestyle, zorder=-10)
    ax.scatter(alpha_escape_keV, y[0], marker='|', color=edgecolor) 
    ax.annotate(element, [right_x, y[0]], xytext=[5, -1], fontsize=8, color='k', 
                textcoords='offset points', ha='left', va='center') 
   
    return ax 


def get_attenuation(element, emission_energy_keV, excitation_energy_keV='rhodium_Ka', h_mm=0.1): 
    '''Attenuation factors for emissions at `emission_keVs` for pure sample with thickness `h_mm`. 

    See Thomsen_2007 equation 11. 
    
    Returns: att_list (array of floats)'''

    if excitation_energy_keV == 'rhodium_Ka': 
        excitation_energy_keV = 22.1 
        
    excitation_energy = excitation_energy_keV * 1000 
    emission_energy = emission_energy_keV * 1000 

    if emission_energy < 150: # eV 
        attenuation = 0 
        
    else: 
        mu_excit = xraydb.mu_elam(element, excitation_energy)
        mu_emiss = xraydb.mu_elam(element, emission_energy)
    
        # mathematically the thickness factor is defined as: 
        rho = xraydb.atomic_density(element)
        thickness_factor = 1 - np.exp(-rho * h_mm * 0.1 * (mu_excit + mu_emiss))
        
        attenuation = (mu_excit / (mu_excit + mu_emiss)) * thickness_factor 

    return attenuation





"""From peak energies to the Periodic Table"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/03_moseley-law.ipynb.

# %% auto 0
__all__ = ['moseley_law']

# %% ../notebooks/03_moseley-law.ipynb 11
import numpy as np 

# %% ../notebooks/03_moseley-law.ipynb 12
def moseley_law(E_K_alpha_keV): 
    '''Square root form of Moseley's law. 
    
    Args: 
        E_K_alfa_keV (float or array of floats): K_alpha peak energy in keV 
        
    Example: 
        moseley_law(6.40)
        
    Returns: 
        Z (float or array of floats): predicted atomic number 
        
    ''' 
    Z = 1 + np.sqrt(1000 * E_K_alpha_keV / 10.2) 
    
    return Z 



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'master',
                'doc_baseurl': '/moseley',
                'doc_host': 'https://fligt.github.io',
                'git_url': 'https://github.com/fligt/moseley',
                'lib_path': 'moseley'},
  'syms': { 'moseley.moseley_law': {'moseley.moseley_law.moseley_law': ('moseley-law.html#moseley_law', 'moseley/moseley_law.py')},
            'moseley.mplot': { 'moseley.mplot.XFluo': ('moseley-law.html#xfluo', 'moseley/mplot.py'),
                               'moseley.mplot.XFluo.__init__': ('moseley-law.html#xfluo.__init__', 'moseley/mplot.py'),
                               'moseley.mplot.XFluo.plot': ('moseley-law.html#xfluo.plot', 'moseley/mplot.py'),
                               'moseley.mplot._colorize': ('moseley-law.html#_colorize', 'moseley/mplot.py'),
                               'moseley.mplot._draw_box': ('moseley-law.html#_draw_box', 'moseley/mplot.py'),
                               'moseley.mplot.make_ptable': ('moseley-law.html#make_ptable', 'moseley/mplot.py'),
                               'moseley.mplot.moseley_law': ('moseley-law.html#moseley_law', 'moseley/mplot.py'),
                               'moseley.mplot.moseley_plot': ('moseley-law.html#moseley_plot', 'moseley/mplot.py'),
                               'moseley.mplot.ptable_plot': ('moseley-law.html#ptable_plot', 'moseley/mplot.py')},
            'moseley.peak_pattern_atlas': { 'moseley.peak_pattern_atlas.PeakPatternAtlas': ( 'a-peak-pattern-atlas.html#peakpatternatlas',
                                                                                             'moseley/peak_pattern_atlas.py'),
                                            'moseley.peak_pattern_atlas.PeakPatternAtlas.__init__': ( 'a-peak-pattern-atlas.html#peakpatternatlas.__init__',
                                                                                                      'moseley/peak_pattern_atlas.py'),
                                            'moseley.peak_pattern_atlas.PeakPatternAtlas.plot_atlas': ( 'a-peak-pattern-atlas.html#peakpatternatlas.plot_atlas',
                                                                                                        'moseley/peak_pattern_atlas.py'),
                                            'moseley.peak_pattern_atlas.PeriodicTable': ( 'a-peak-pattern-atlas.html#periodictable',
                                                                                          'moseley/peak_pattern_atlas.py'),
                                            'moseley.peak_pattern_atlas.PeriodicTable.__init__': ( 'a-peak-pattern-atlas.html#periodictable.__init__',
                                                                                                   'moseley/peak_pattern_atlas.py'),
                                            'moseley.peak_pattern_atlas.PeriodicTable.plot_table': ( 'a-peak-pattern-atlas.html#periodictable.plot_table',
                                                                                                     'moseley/peak_pattern_atlas.py'),
                                            'moseley.peak_pattern_atlas._draw_box': ( 'a-peak-pattern-atlas.html#_draw_box',
                                                                                      'moseley/peak_pattern_atlas.py'),
                                            'moseley.peak_pattern_atlas._fetch_table': ( 'a-peak-pattern-atlas.html#_fetch_table',
                                                                                         'moseley/peak_pattern_atlas.py'),
                                            'moseley.peak_pattern_atlas._sort_ptrn_list': ( 'a-peak-pattern-atlas.html#_sort_ptrn_list',
                                                                                            'moseley/peak_pattern_atlas.py'),
                                            'moseley.peak_pattern_atlas.colorize': ( 'a-peak-pattern-atlas.html#colorize',
                                                                                     'moseley/peak_pattern_atlas.py')},
            'moseley.xrf_physics': { 'moseley.xrf_physics.ElementXRF': ( 'theoretical-peak-patterns.html#elementxrf',
                                                                         'moseley/xrf_physics.py'),
                                     'moseley.xrf_physics.ElementXRF.__init__': ( 'theoretical-peak-patterns.html#elementxrf.__init__',
                                                                                  'moseley/xrf_physics.py'),
                                     'moseley.xrf_physics.ElementXRF.plot_spectrum': ( 'theoretical-peak-patterns.html#elementxrf.plot_spectrum',
                                                                                       'moseley/xrf_physics.py'),
                                     'moseley.xrf_physics.find_peaks': ( 'theoretical-peak-patterns.html#find_peaks',
                                                                         'moseley/xrf_physics.py'),
                                     'moseley.xrf_physics.gaussian_convolve': ( 'theoretical-peak-patterns.html#gaussian_convolve',
                                                                                'moseley/xrf_physics.py'),
                                     'moseley.xrf_physics.get_attenuation': ( 'theoretical-peak-patterns.html#get_attenuation',
                                                                              'moseley/xrf_physics.py'),
                                     'moseley.xrf_physics.plot_pattern': ( 'theoretical-peak-patterns.html#plot_pattern',
                                                                           'moseley/xrf_physics.py')}}}
